import java.io.*;

/**
 * Class which can write individual bits to an output stream.
 * 
 * @author Amanda Askew & Dane Barney
 * @version 2003.09.11
 */
public class BitWriter {
    public DataOutputStream out;
    private int numBits;
    private byte buffer;
    public int bytesWritten;
    
    public int num0s;
    public int num1s;

    /**
     * Initializes the output stream.
     */
    public BitWriter(DataOutputStream out) {
        this.out = out;
        buffer = 0;
        numBits = 0;
        bytesWritten = 0;
    }
    
    public void close() throws IOException {
        out.close();
    }
    
    /**
     * Writes the remaining incomplete buffer to file (adds 0's to fill unused bit places)
     */
    public void finish() throws IOException {
        if(numBits != 0) {
            buffer <<= 8 - numBits;
            out.writeByte(buffer);
            bytesWritten++;
            numBits = 0;
            buffer = 0;
        }
    }

    /**
     * Adds a bit to the bitstream, and writes to file if the bitstream
     * contains a full byte.
     */
    public void writeBit(int b) throws IOException {
        assert (b == 0 || b == 1);
        if (b==0) num0s++;
        if (b==1) num1s++;
        buffer <<= 1;
        buffer |= b;
        numBits++;
        if(numBits == 8) {
            out.writeByte(buffer);
            bytesWritten++;
            buffer = 0;
            numBits = 0;
        }
        assert (numBits >= 0);
        assert (numBits <= 7);
    }

    /**
     * Write 4 bits
     */
    public void writeNibble(int b) throws IOException {
        assert (b >= 0 && b <= 15);
        byte temp;
        for(int shiftBits=3; shiftBits >= 0; shiftBits--) {
            temp = (byte)((b >>> shiftBits) & 1);
            writeBit(temp);
        }
    }

    public void writeByte(byte b) throws IOException {
        if(numBits == 0) {
            out.writeByte(b);
            bytesWritten++;
        } else {
            byte temp;
            for(int shiftBits=7; shiftBits >= 0; shiftBits--) {
                temp = (byte)((b >>> shiftBits) & 1);
                writeBit(temp);
            }
        }
    }

    public void write2Bytes(short b) throws IOException {
        if(numBits == 0) {
            out.writeShort(b);
            bytesWritten += 2;
        } else {
            byte temp;
            for(int shiftBits=15; shiftBits >= 0; shiftBits--) {
                temp = (byte)((b >>> shiftBits) & 1);
                writeBit(temp);
            }
        }
    }

    public void write3Bytes(int b) throws IOException {
        assert (b >= 0 && b <= 16777215);
        if(numBits == 0) {
            out.writeShort((short)(b >>> 8));
            out.writeByte((byte)b);
            bytesWritten += 3;
        } else {
            byte temp;
            for(int shiftBits=23; shiftBits >= 0; shiftBits--) {
                temp = (byte)((b >>> shiftBits) & 1);
                writeBit(temp);
            }
        }
    }

    public void write4Bytes(int b) throws IOException {
        if(numBits == 0) {
            out.writeInt(b);
            bytesWritten += 4;
        } else {
            byte temp;
            for(int shiftBits=31; shiftBits >= 0; shiftBits--) {
                temp = (byte)((b >>> shiftBits) & 1);
                writeBit(temp);
            }
        }
    }
    
    public void writeFloat(float f) throws IOException {
        out.writeFloat(f);
        bytesWritten += 4;
    }

    public void writeDouble(double d) throws IOException {
        out.writeDouble(d);
        bytesWritten += 8;
    }
}
