/**
 * Defines a context to be used during the arithmetic coding algorithm.
 * 
 * @author Amanda Askew & Dane Barney
 * @version 2003.09.11
 */
public class Context {
    public int[] count; // the count of each value
    public int[] cum_count; // the boundary on the interval for each value
    public int total_count;
    public int indexNum; // for debugging purposes

    /**
     * Creates a new context with initial counts of values
     * 
     * @param count  the initial count for each value
     */
    public Context(int[] count) {
        this.count = count;
        cum_count = new int[count.length+1]; // 1 slot larger because cum_count[0] doesn't correspond to a value (it's always 0)
        
        // initialize cum_count array based on counts
        int total = 0;
        for(int i=0; i < count.length; i++) {
            total += count[i];
            cum_count[i+1] = total; // cum_count[i+1] = value in count[i] because cum_count[0] doesn't correspond to a value
        }
        
        total_count = total;
    }
  
    /* 
     * Increments the count for this particular value, and updates all variables.
     */
    public void incrementCount(int val) {
        count[val]++;
        
        // update all values in cum_count array affected by this increase
        for(int i=val; i < this.count.length; i++)
            cum_count[i+1]++;
        
        total_count++;
    }
}