import java.io.*;

public class DiscreteCosineTransform {
    private int oldtogether = -1;

    public int blockSize;     // 8 or 4 (for 8x8 or 4x4 blocks)
    public int nSubbands;     // # of subbands in transform
    public double upperBound;    // the smallest power of 2 greater than the largest abs val coefficient
    public double DCavg;
    public int hsize;
    public int vsize;



    // pointers to start of subbands
    public int subbandSize;    // # of coeffs in each subband  (equivalent to number of blocks)
    public int subbandHsize;   // horizontal size of subband
    public int subbandVsize;   // vertical size of subband
    public double[][] subbandPtr; // pointer to start of each subband [subbandNum][blockNum]
    

    public static void main(String[] args) throws InvalidFormatException, IOException {
	Image im = new Image(args[0], true);
	DiscreteCosineTransform dct = new DiscreteCosineTransform (8, im.height, im.width);

	// display image
	for (int i = 0; i < im.height; i+=1){
	    for (int j = 0; j < im.width; j+=1){
		System.out.print(im.pixels[j][i] + " ");
	    }
	    System.out.println();
	}

	dct.forwardTransform(im);
	int[][] inverted = dct.inverseTransform();

	System.out.println();
	for (int i = 0; i < im.height; i+=1){
	    for (int j = 0; j < im.width; j+=1){
		System.out.print(inverted[j][i] + " ");
	    }
	    System.out.println();
	}	

    }


    public DiscreteCosineTransform(int blockSize, int hsize, int vsize) throws InvalidFormatException {
	this.blockSize = blockSize;
	this.nSubbands = blockSize * blockSize;
	
	init(hsize,vsize);
	System.out.println("new DCT! hsize = " + hsize +"; vsize = " + vsize);
    }
    
    public void forwardTransform (Image image) throws InvalidFormatException {
	if ((image.width%blockSize != 0)|| (image.height%blockSize != 0)){
	    throw new InvalidFormatException("image hieght and width must be a multiple of blocksize");
	}

	int[][] pixels = image.pixels;

	double[][][] transformedBlocks = new double[subbandSize][][];

		
	for (int bh = 0; bh < image.height; bh+= blockSize){
	    for (int bw = 0; bw < image.width; bw += blockSize){
		
		//get original NxN block
		int[][] originalBlock = new int[blockSize][blockSize];
		for (int h = 0; h < blockSize; h += 1){
		    for (int w = 0; w < blockSize; w += 1){
			originalBlock[w][h] = pixels[bw+w][bh+h];
		    }
		}

		double[][] transformedBlock = new double[blockSize][blockSize];
		// performing discrete cosine transform on a block

		/* TODO: the next 5 lines currently copies the original block into transformedBlock.  
		 * You need to actually perform DCT on the originalBlock 
		 * and place the result in transformedBlock
		 */
		for (int i = 0; i < blockSize; i += 1){
		    for (int j = 0; j < blockSize; j += 1){
			transformedBlock[i][j] = (double)originalBlock[i][j];
		    }
		}
				
		transformedBlocks[(bh/blockSize)*subbandHsize + (bw/blockSize)] = transformedBlock; 
		
	    }
	}
	
        // linearize data
	blocksToSubBands(transformedBlocks);
    }

    // fills subBandPtr matrix so that we can access the transformed blocks by subband
    private void blocksToSubBands(double[][][] transformedBlocks){
	
	for (int sb = 0; sb < nSubbands; sb += 1){
	    for (int bn = 0; bn < subbandSize; bn += 1) {
		subbandPtr[sb][bn] = transformedBlocks[bn][sb/blockSize][sb%blockSize];
		// possible bug:  may have to switch [sb/blockSize] and [sb%blockSize]
	    }
	}	
    }

    private double[][][] subBandsToBlocks(){
	double[][][] transformedBlocks = new double[subbandSize][blockSize][blockSize];
	for (int sb = 0; sb < subbandSize; sb += 1){
	    for (int h = 0; h < blockSize; h += 1){
		for (int w = 0; w < blockSize; w += 1){
		    transformedBlocks[sb][w][h] = subbandPtr[h*blockSize+w][sb];
		}
	    }
	}
	return transformedBlocks;
    }

    public int[][] inverseTransform() {
        double[][][] transformed = subBandsToBlocks();
	//blocks are [blockNum][w][h]
	
	double[][][] invertedImage = new double[subbandSize][][];

	for (int blockNum = 0; blockNum < subbandSize; blockNum += 1){
	    double[][] invertedBlock = new double[blockSize][blockSize];

	    // performing inverse discrete cosine transform on a block
	    /* TODO: the next line currently copies the transformed block to the inverted block.
	     * You need to actually implement the inverse DCT so that 
	     * the inverse transform is performed on transformed[blockNum] 
	     * and the result is saved in invertedBlock
	     */
	    invertedBlock = transformed[blockNum];
	    
	    invertedImage[blockNum] = invertedBlock;
	}


        int[][] roundedImage = new int[hsize][vsize];

	for (int blockNum = 0; blockNum < subbandSize; blockNum += 1){
	    for (int n = 0; n < blockSize; n += 1){
		for (int m = 0; m < blockSize; m += 1){
		    int i = ((blockNum%subbandHsize)*blockSize) + m;  // width
		    int j = ((blockNum/subbandHsize)*blockSize) + n; // height
		    roundedImage[i][j] = (int)Math.round(invertedImage[blockNum][n][m]);
		}
	    }
	}
        
        return roundedImage;
    }
    
	
    public void subtractDCavg() {
        // Find average coeff value in DC subband and store it
        for(int i=0; i < subbandSize; i++)
	    DCavg += subbandPtr[0][i];
        DCavg /= subbandSize;

        // Subtract it off all DC coeffs
        for(int i=0; i < subbandSize; i++)
            subbandPtr[0][i] -= DCavg;

	System.out.println("DCavg1: " + DCavg);
    }
    
    public void addDCavg() {
        // Add average coeff value in DC subband to coeffs
        for(int i=0; i < subbandSize; i++)
            subbandPtr[0][i] += DCavg;
    }
	
    
    public void getUpperBound() {
        // Find maximum absolute value coeff
        double maxCoeff = 0;
        for(int i=0; i < nSubbands; i++)
            for(int j=0; j < subbandSize; j++)
                if(Math.abs(subbandPtr[i][j]) > maxCoeff) maxCoeff = Math.abs(subbandPtr[i][j]);
        
        // Find smallest power of 2 greater than or equal to maxCoeff
        double temp = 0;
        for(int i=0; temp < maxCoeff; i++)
            temp = Math.pow(2, i);
        upperBound = temp;

	System.out.println("upperBound1: " + upperBound);
    }

        

    private void init(int hsize, int vsize){
	this.hsize = hsize;
	this.vsize = vsize;

	subbandHsize = hsize/blockSize;
        subbandVsize = vsize/blockSize;
	System.out.println("just set subbandHsize to " + subbandHsize);
	System.out.println("just set subbandVsize to " + subbandVsize);
	subbandSize = subbandHsize * subbandVsize; // equal to number of blocks
        subbandPtr = new double[nSubbands][];

        for(int i=0; i < nSubbands; i++) {
	    subbandPtr[i] = new double[subbandSize];
        }
    }

}
