/**
 * Implements a doubly linked list data structure of ListNode objects.
 */
public class DoublyLinkedList {
    private ListNode head;
    private ListNode tail;
    public int size;

    public DoublyLinkedList() {
        head = new ListNode(null);
        tail = head;
        size = 0;
    }

    public boolean isEmpty() {
        return this.head.next == null;
    }
    
    public void makeEmpty() {
        head.next = null;
        tail = head;
        size = 0;
    }
    
    /**
     * Attach another linked list to the end of this one.
     */
    public void append(DoublyLinkedList l) {
        if(l.head.next != null) {
            tail.next = l.head.next;
            tail.next.prev = tail;
            tail = l.tail;
            size += l.size;
        }
    }
    
    /**
     * Insert at the end of the list.
     * @param n the item to insert.
     */
    public void insert(ListNode n) {
        tail.next = n;
        n.prev = tail;
        tail = n;
        n.next = null;
        n.inList = this;
        size++;
    }

    /**
     * Remove and return the first item.
     */
    public ListNode remove() {
        ListNode n = null;
        if(head.next != null) { // if list is not empty
            size--;
            if(head.next.equals(tail)) // if list contains only 1 node
                tail = head;
            n = head.next;
            head.next = head.next.next;
            if(head.next != null)
                head.next.prev = head;
            n.inList = null;
        }
        assert (size >= 0);
        return n;
    }
    
    /**
     * Remove and return the first occurrence of an item.
     * @param op the item to remove.
     */
    public ListNode remove(ListNode n) {
        assert this.contains(n);
        if(n.equals(tail)) // if r is last node
            tail = n.prev;
        n.prev.next = n.next; // bypass deleted node
        if(n.next != null)
            n.next.prev = n.prev;
        size--;
        assert (size >= 0);
        return n;
    }
    
    public boolean contains(ListNode n) {
        return this.equals(n.inList);
    }
}