import java.awt.image.BufferedImage;
import java.awt.image.Raster;
import java.io.DataInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.Iterator;

import javax.imageio.IIOImage;
import javax.imageio.ImageIO;
import javax.imageio.ImageReader;
import javax.imageio.ImageWriteParam;
import javax.imageio.ImageWriter;
import javax.imageio.stream.FileImageInputStream;
import javax.imageio.stream.FileImageOutputStream;
import javax.imageio.ImageTypeSpecifier;
import javax.imageio.ImageReadParam;
public class JPEG {

	public static void main(String[] args)
		throws FileNotFoundException, IOException, InvalidFormatException {
		String input = args[0];
		String output = args[1];
		float bpp = Float.valueOf(args[2]).floatValue();

		File inFile = new File(input);
		FileImageInputStream inStream = new FileImageInputStream(inFile);
		Iterator readers = ImageIO.getImageReaders(inStream);
		if (!readers.hasNext())
			throw new InvalidFormatException("Unrecognized input image format.");
		ImageReader reader = (ImageReader)readers.next();
		reader.setInput(inStream, true);
		//if (reader.getNumImages(true) > 1)
		//throw new InvalidFormatException("This file contains multiple embedded images.");

		System.err.println(
			"Input: "
				+ input
				+ "\n"
				+ "\t   format: "
				+ reader.getOriginatingProvider().getMIMETypes()[0]
				+ "\n"
				+ "\t    width: "
				+ reader.getWidth(0)
				+ "\n"
				+ "\t   height: "
				+ reader.getHeight(0)
				+ "\n"
				+ "\t filesize: "
				+ inFile.length()
				+ " bytes\n");

		ImageReadParam irp = new ImageReadParam();
		irp.setDestinationType(ImageTypeSpecifier.createFromBufferedImageType(BufferedImage.TYPE_BYTE_GRAY ));

		BufferedImage bi = reader.read(0, irp);
		inStream.close();
		System.out.println("bi.getType(): " + bi.getType());
		System.out.println("bi.getRaster().getNumBands(): " + bi.getRaster().getNumBands());
		
		if (bi.getType() != BufferedImage.TYPE_BYTE_GRAY)
		    throw new InvalidFormatException("This is not a grayscale image.");
		assert bi.getRaster().getNumBands() == 1;

		// initialize output stream based on specified output format
		Iterator writers = ImageIO.getImageWritersByFormatName("jpeg");
		ImageWriter writer = (ImageWriter)writers.next();
		ImageWriteParam param = writer.getDefaultWriteParam();
		param.setCompressionMode(ImageWriteParam.MODE_EXPLICIT);
		param.setCompressionType(param.getCompressionTypes()[0]);
		IIOImage outImg = new IIOImage(bi, null, null);
		File outFile = new File(output);

		// set output bitrate
		double minQual = 0;
		double maxQual = 1;
		double currQual, currBpp;
		do {
			outFile.delete();
			FileImageOutputStream outStream =
				new FileImageOutputStream(outFile);
			writer.setOutput(outStream);

			currQual = (minQual + maxQual) / 2;
			param.setCompressionQuality((float)currQual);
			writer.write(null, outImg, param);
			outStream.close();

			DataInputStream in =
				new DataInputStream(new FileInputStream(outFile));

			int headerSize = 0;
			int b = in.readUnsignedShort();
			headerSize += 2;
			assert b == 0xFFD8;
			b = in.readUnsignedShort();
			while (b == 0xFFE0
				|| b == 0xFFDB
				|| b == 0xFFC0
				|| b == 0xFFC4
				|| b == 0xFFDA) {
				headerSize += 2;
				b = in.readUnsignedShort();
				in.skipBytes(b - 2);
				headerSize += b;
				b = in.readUnsignedShort();
			}
			headerSize += 2; // for terminating 0xFFD9
			in.close();

			long imageData = outFile.length() - headerSize;
			currBpp =
				(double)imageData * 8.0 / (bi.getWidth() * bi.getHeight());
			System.err.println(
				"quality = "
					+ currQual
					+ "\theader = "
					+ headerSize
					+ "\timage = "
					+ imageData
					+ "\tbitrate = "
					+ currBpp);

			if ((maxQual - minQual) < 0.00000001)
				break;

			if (currBpp < bpp)
				minQual = currQual;
			else
				maxQual = currQual;

			writer.reset();

		}
		while (Math.abs(currBpp - bpp) > 0.00000001);

		BufferedImage bi2 = ImageIO.read(outFile);
		double psnr = printPSNR(bi, bi2);
		System.out.println(currBpp+"\t"+psnr);
	}

	/**
	 * Computes and prints out the PSNR of two images (which must have the same 
	 * dimensions and type).
	 */
	public static double printPSNR(BufferedImage im1, BufferedImage im2) {
		assert(
			im1.getType() == im2.getType()
				&& im1.getHeight() == im2.getHeight()
				&& im1.getWidth() == im2.getWidth());

		double mse = 0;
		int width = im1.getWidth();
		int height = im1.getHeight();
		Raster r1 = im1.getRaster();
		Raster r2 = im2.getRaster();
		for (int j = 0; j < height; j++)
			for (int i = 0; i < width; i++)
				mse
					+= Math.pow(r1.getSample(i, j, 0) - r2.getSample(i, j, 0), 2);
		mse /= (double) (width * height);
		System.err.println("MSE = " + mse);
		int maxVal = 255;
		double x = Math.pow(maxVal, 2) / mse;
		double psnr = 10.0 * logbase10(x);
		System.err.println("PSNR = " + psnr);
		return psnr;
	}

	/**
	 * Returns the base-10 logarithm of a number
	 */
	public static double logbase10(double x) {
		return Math.log(x) / Math.log(10);
	}

}
