import java.awt.image.BufferedImage;
import java.awt.image.Raster;
import java.io.DataInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.Iterator;

import javax.imageio.IIOImage;
import javax.imageio.ImageIO;
import javax.imageio.ImageReader;
import javax.imageio.ImageWriteParam;
import javax.imageio.ImageWriter;
import javax.imageio.stream.FileImageInputStream;
import javax.imageio.stream.FileImageOutputStream;

import com.sun.media.imageio.plugins.jpeg2000.J2KImageWriteParam;

public class JPEG2000 {

	public static void main(String[] args)
		throws FileNotFoundException, IOException, InvalidFormatException {
		String input = args[0];
		String output = args[1];
		double bpp = Double.valueOf(args[2]).doubleValue();

		File inFile = new File(input);
		FileImageInputStream inStream = new FileImageInputStream(inFile);
		Iterator readers = ImageIO.getImageReaders(inStream);
		if (!readers.hasNext())
			throw new InvalidFormatException("Unrecognized input image format.");
		ImageReader reader = (ImageReader)readers.next();
		reader.setInput(inStream, true);
		//if (reader.getNumImages(true) > 1)
		//throw new InvalidFormatException("This file contains multiple embedded images.");

		System.err.println(
			"Input: "
				+ input
				+ "\n"
				+ "\t   format: "
				+ reader.getOriginatingProvider().getMIMETypes()[0]
				+ "\n"
				+ "\t    width: "
				+ reader.getWidth(0)
				+ "\n"
				+ "\t   height: "
				+ reader.getHeight(0)
				+ "\n"
				+ "\t filesize: "
				+ inFile.length()
				+ " bytes\n");

		BufferedImage bi = reader.read(0);
		inStream.close();
		//if (bi.getType() != BufferedImage.TYPE_BYTE_GRAY)
		//throw new InvalidFormatException("This is not a grayscale image.");
		assert bi.getRaster().getNumBands() == 1;

		// initialize output stream based on specified output format
		Iterator writers = ImageIO.getImageWritersByFormatName("jpeg2000");
		ImageWriter writer = (ImageWriter)writers.next();
		J2KImageWriteParam param = (J2KImageWriteParam)writer.getDefaultWriteParam();
		param.setLossless(false);
		param.setCompressionMode(ImageWriteParam.MODE_EXPLICIT);

	

		param.setCompressionType(param.getCompressionTypes()[0]);
		
		System.out.println("param.getCompressionType(): " + param.getCompressionType());	

		param.setEncodingRate(bpp);

		System.out.println("bpp: " + param.getEncodingRate());
		IIOImage outImg = new IIOImage(bi, null, null);
		File outFile = new File(output);

		// set output bitrate
		outFile.delete();
		FileImageOutputStream outStream = new FileImageOutputStream(outFile);
		writer.setOutput(outStream);

		writer.write(null, outImg, param);
		outStream.close();

		BufferedImage bi2 = ImageIO.read(outFile);
		double psnr = printPSNR(bi, bi2);
		System.out.println(bpp+"\t"+psnr);
	}

	/**
	 * Computes and prints out the PSNR of two images (which must have the same 
	 * dimensions and type).
	 */
	public static double printPSNR(BufferedImage im1, BufferedImage im2) {
		assert(
			im1.getType() == im2.getType()
				&& im1.getHeight() == im2.getHeight()
				&& im1.getWidth() == im2.getWidth());

		double mse = 0;
		int width = im1.getWidth();
		int height = im1.getHeight();
		Raster r1 = im1.getRaster();
		Raster r2 = im2.getRaster();
		for (int j = 0; j < height; j++)
			for (int i = 0; i < width; i++)
				mse
					+= Math.pow(r1.getSample(i, j, 0) - r2.getSample(i, j, 0), 2);
		mse /= (double) (width * height);
		System.err.println("MSE = " + mse);
		int maxVal = 255;
		double x = Math.pow(maxVal, 2) / mse;
		double psnr = 10.0 * logbase10(x);
		System.err.println("PSNR = " + psnr);
		return psnr;
	}

	/**
	 * Returns the base-10 logarithm of a number
	 */
	public static double logbase10(double x) {
		return Math.log(x) / Math.log(10);
	}

}
